/*
 * Korseby Fan Adjustment Daemon
 * Copyright (C) 2009 Kristian Peters, Korseby
 *
 * note: loosely based on FanCantrol 1.2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <IOKit/IOKitLib.h>

#include "kfand.h"

static io_connect_t conn;



UInt32 _strtoul(char *str, int size, int base)
{
	UInt32 total = 0;
	int i;
	
	for (i = 0; i < size; i++) {
		if (base == 16)
			total += str[i] << (size - 1 - i) * 8;
		else
		total += (unsigned char) (str[i] << (size - 1 - i) * 8);
	}
	return total;
}



void _ultostr(char *str, UInt32 val)
{
	str[0] = '\0';
	sprintf(str, "%c%c%c%c", 
			(unsigned int) val >> 24,
			(unsigned int) val >> 16,
			(unsigned int) val >> 8,
			(unsigned int) val);
}



float _strtof(char *str, int size, int e)
{
	float total = 0;
	int i;
	
	for (i = 0; i < size; i++) {
		if (i == (size - 1))
			total += (str[i] & 0xff) >> e;
		else
 			total += str[i] << (size - 1 - i) * (8 - e);
	}
	
	return total;
}



void printFPE2(SMCVal_t val)
{
	/* FIXME: This decode is incomplete, last 2 bits are dropped */
	printf("%.0f ", _strtof(val.bytes, val.dataSize, 2));
}



void printUInt(SMCVal_t val)
{
	printf("%u ", (unsigned int) _strtoul(val.bytes, val.dataSize, 10));
}



void printBytesHex(SMCVal_t val)
{
	int i;
	
	printf("(bytes");
	for (i = 0; i < val.dataSize; i++)
		printf(" %02x", (unsigned char) val.bytes[i]);
	printf(")\n");
}



void printVal(SMCVal_t val)
{
	printf("  %-4s  [%-4s]  ", val.key, val.dataType);
	
	if (val.dataSize > 0) {
		if ((strcmp(val.dataType, DATATYPE_UINT8) == 0) ||
			(strcmp(val.dataType, DATATYPE_UINT16) == 0) ||
			(strcmp(val.dataType, DATATYPE_UINT32) == 0))
				printUInt(val);
		else if (strcmp(val.dataType, DATATYPE_FPE2) == 0)
				printFPE2(val);
		printBytesHex(val);
	}
	else {
			printf("no data\n");
	}
}



kern_return_t SMCOpen(void)
{
	kern_return_t	result;
	mach_port_t		masterPort;
	io_iterator_t	iterator;
	io_object_t		device;
	
	result = IOMasterPort(MACH_PORT_NULL, &masterPort);
	
	CFMutableDictionaryRef matchingDictionary = IOServiceMatching("AppleSMC");
	result = IOServiceGetMatchingServices(masterPort, matchingDictionary, &iterator);
	if (result != kIOReturnSuccess)
	{
		printf("Error: IOServiceGetMatchingServices() = %08x\n", result);
		return 1;
	}
	
	device = IOIteratorNext(iterator);
	IOObjectRelease(iterator);
	if (device == 0)
	{
		printf("Error: no SMC found\n");
		return 1;
	}
	
	result = IOServiceOpen(device, mach_task_self(), 0, &conn);
	IOObjectRelease(device);
	if (result != kIOReturnSuccess)
	{
		printf("Error: IOServiceOpen() = %08x\n", result);
		return 1;
	}
	
	return kIOReturnSuccess;
}



kern_return_t SMCClose()
{
	return IOServiceClose(conn);
}



kern_return_t SMCCall(int index, SMCKeyData_t *inputStructure, SMCKeyData_t *outputStructure)
{
	IOItemCount		structureInputSize;
	IOByteCount		structureOutputSize;
	
	structureInputSize = sizeof(SMCKeyData_t);
	structureOutputSize = sizeof(SMCKeyData_t);
	
	return IOConnectMethodStructureIStructureO(
			conn,
			index,
			structureInputSize,
			&structureOutputSize,
			inputStructure,
			outputStructure
	);
}



kern_return_t SMCReadKey(UInt32Char_t key, SMCVal_t *val)
{
	kern_return_t	result;
	SMCKeyData_t	inputStructure;
	SMCKeyData_t	outputStructure;
	
	memset(&inputStructure, 0, sizeof(SMCKeyData_t));
	memset(&outputStructure, 0, sizeof(SMCKeyData_t));
	memset(val, 0, sizeof(SMCVal_t));
	
	inputStructure.key = _strtoul(key, 4, 16);
	sprintf(val->key, key);
	inputStructure.data8 = SMC_CMD_READ_KEYINFO;	
	
	result = SMCCall(KERNEL_INDEX_SMC, &inputStructure, &outputStructure);
	
	if (result != kIOReturnSuccess)
		return result;
	
	val->dataSize = outputStructure.keyInfo.dataSize;
	_ultostr(val->dataType, outputStructure.keyInfo.dataType);
	inputStructure.keyInfo.dataSize = val->dataSize;
	inputStructure.data8 = SMC_CMD_READ_BYTES;
	
	result = SMCCall(KERNEL_INDEX_SMC, &inputStructure, &outputStructure);
	
	if (result != kIOReturnSuccess)
		return result;
	
	memcpy(val->bytes, outputStructure.bytes, sizeof(outputStructure.bytes));
	
	return kIOReturnSuccess;
}



kern_return_t SMCWriteKey(SMCVal_t writeVal)
{
	kern_return_t	result;
	SMCKeyData_t	inputStructure;
	SMCKeyData_t	outputStructure;
	
	SMCVal_t		readVal;
	
	result = SMCReadKey(writeVal.key, &readVal);
	
	if (result != kIOReturnSuccess) 
		return result;
	
	if (readVal.dataSize != writeVal.dataSize)
		return kIOReturnError;
	
	memset(&inputStructure, 0, sizeof(SMCKeyData_t));
	memset(&outputStructure, 0, sizeof(SMCKeyData_t));
	
	inputStructure.key = _strtoul(writeVal.key, 4, 16);
	inputStructure.data8 = SMC_CMD_WRITE_BYTES;	
	inputStructure.keyInfo.dataSize = writeVal.dataSize;
	
	memcpy(inputStructure.bytes, writeVal.bytes, sizeof(writeVal.bytes));
	
	result = SMCCall(KERNEL_INDEX_SMC, &inputStructure, &outputStructure);
	
	if (result != kIOReturnSuccess)
		return result;
	
	return kIOReturnSuccess;
}



kern_return_t SMCPrintFans(void)
{
	kern_return_t	result;
	SMCVal_t		val;
	UInt32Char_t	key;
	int				totalFans, i;
	
	result = SMCReadKey("FNum", &val);
	
	if (result != kIOReturnSuccess)
		return kIOReturnError;
	
	totalFans = _strtoul(val.bytes, val.dataSize, 10);
	SMCFansAmount = totalFans;
	
	for (i = 0; i < totalFans; i++) {
		printf("\nFan %d/%d:\n", i+1, totalFans);
		sprintf(key, "F%dAc", i);
		
		SMCReadKey(key, &val);
		printf("	Actual speed  : %.0f\n", _strtof(val.bytes, val.dataSize, 2));
		sprintf(key, "F%dMn", i);
		
		SMCReadKey(key, &val);
		printf("	Minimum speed : %.0f\n", _strtof(val.bytes, val.dataSize, 2));
		sprintf(key, "F%dMx", i);
		
		SMCReadKey(key, &val);
		printf("	Maximum speed : %.0f\n", _strtof(val.bytes, val.dataSize, 2));
		sprintf(key, "F%dSf", i);
		
		SMCReadKey(key, &val);
		printf("	Safe speed    : %.0f\n", _strtof(val.bytes, val.dataSize, 2));
		sprintf(key, "F%dTg", i);
		
		SMCReadKey(key, &val);
		printf("	Target speed  : %.0f\n", _strtof(val.bytes, val.dataSize, 2));
		SMCReadKey("FS! ", &val);
		
		if ((_strtoul(val.bytes, 2, 16) & (1 << i)) == 0)
			printf("	Mode          : auto\n\n"); 
		else
			printf("	Mode          : forced\n\n");
	}
	
	return kIOReturnSuccess;
}



double SMCGetTemperature(char *key)
{
	SMCVal_t		val;
	kern_return_t	result;
	
	result = SMCReadKey(key, &val);
	
	if (result == kIOReturnSuccess) {
		if (val.dataSize > 0) {
			if (strcmp(val.dataType, DATATYPE_SP78) == 0) {
				/* convert fp78 value to temperature */
				int intValue = (val.bytes[0] * 256 + val.bytes[1]) >> 2;
				return intValue / 64.0;
			}
		}
	}
	
	return 0.0;
}



kern_return_t SMCSetFanRpm(char *key, int rpm)
{
	SMCVal_t		val;
	
	strcpy(val.key, key);
	
	val.bytes[0] = (rpm << 2) / 256;
	val.bytes[1] = (rpm << 2) % 256;
	val.dataSize = 2;
	
	return SMCWriteKey(val);
}



int SMCGetFanRpm(char *key)
{
	SMCVal_t		val;
	kern_return_t	result;
	
	result = SMCReadKey(key, &val);
	
	if (result == kIOReturnSuccess) {
		if (val.dataSize > 0) {
			if (strcmp(val.dataType, DATATYPE_FPE2) == 0) {
				/* convert FPE2 value to int value */
				return (int)_strtof(val.bytes, val.dataSize, 2);
			}
		}
	}
	
	return -1;
}



int main(int argc, char *argv[])
{
	pid_t			pid, sid;
	SMCVal_t		val;
	double			temps[3];
	int				rpm_before = RPM_DEFAULT;
	
	/* open smc connection */
	if (SMCOpen() != kIOReturnSuccess) {
		printf("Error: Could not find SMC.");
		return SMC_ERROR_CONNECTION;
	}
	
	/* print info */
	SMCPrintFans();
	
	/* write values */
	strcpy(val.key, "FS! ");
	val.bytes[0] = 0;
	val.bytes[1] = 0;
	val.dataSize = 2;
	
	if (SMCWriteKey(val) != kIOReturnSuccess) {
		printf("Error: Could not write to SMC.\n");
		SMCClose();
		return SMC_ERROR_WRITE;
	}
	
	/* temp history */
	temps[0] = temps[1] = temps[2] = SMCGetTemperature(SMC_KEY_CPU_TEMP);
	
	/* set min rpm */
	if (SMCSetFanRpm(SMC_KEY_FAN0_RPM_MIN, RPM_MIN) != kIOReturnSuccess) {
		printf("Error: Could not write to SMC.\n");
		SMCClose();
		return SMC_ERROR_WRITE_TEMP;
	}
	if (SMCFansAmount > 1)
		SMCSetFanRpm(SMC_KEY_FAN1_RPM_MIN, RPM_MIN);
	
	/* set current rpm */
	SMCSetFanRpm(SMC_KEY_FAN0_RPM_CUR, RPM_DEFAULT);
	if (SMCFansAmount > 1)
		SMCSetFanRpm(SMC_KEY_FAN1_RPM_CUR, RPM_DEFAULT);
	
	SMCClose();
	
	
	/* fork off parent process */
	pid = fork();
	if (pid < 0) {
		exit(EXIT_FAILURE);
	}
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}
	
	/* change file mode mask */
	umask(0);
	
	/* create new sid for child process */
	sid = setsid();
	if (sid < 0) {
		exit(EXIT_FAILURE);
	}
	
	/* log daemonized fork */
	printf("fan daemon is running in background.\n");
	
	/* close out the standard file descriptors */
#ifndef _DEBUG_
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
#endif
	
	
	/* loop */
	while (1) {
		SMCOpen();
		
		double temp = SMCGetTemperature(SMC_KEY_CPU_TEMP);
		
		if (temp != 0.0) {
			temps[0] = temps[1];
			temps[1] = temps[2];
			temps[2] = temp;
			
			/* temp history */
			temp = (temps[0] + temps[1] + temps[2]) / 3.0;
			
			/* set new rpm */			
			int rpm = RPM_DEFAULT + (temp - TEMP_DEFAULT) * (RPM_MAX - RPM_DEFAULT) / (TEMP_MAX - TEMP_DEFAULT);
			
			if (rpm > RPM_MAX)
				rpm = RPM_MAX;
			if (rpm < RPM_DEFAULT)
				rpm = RPM_DEFAULT;
			if (rpm + 200 < rpm_before)			/* slower down scaling */
				rpm = rpm_before - 200;
			if (temp > TEMP_MAX)
				rpm = RPM_MAX;
			
			/* write new rpm to SMC */
			/* note: SMC_KEY_FAN0_RPM_CUR does not work for long, better set RPM_MAX instead */
			if (rpm - RPM_DYNAMIC > rpm_before) {
				SMCSetFanRpm(SMC_KEY_FAN0_RPM_MAX, rpm + RPM_MAX_OFFSET);
				if (SMCFansAmount > 1)
					SMCSetFanRpm(SMC_KEY_FAN1_RPM_MAX, rpm + RPM_MAX_OFFSET);
#ifdef _DEBUG_
			printf("Adjustment up needed (%d-%d). ", rpm, rpm_before);
#endif			
			} else if (rpm + RPM_DYNAMIC <= rpm_before) {
				SMCSetFanRpm(SMC_KEY_FAN0_RPM_MAX, rpm + RPM_MAX_OFFSET);
				if (SMCFansAmount > 1)
					SMCSetFanRpm(SMC_KEY_FAN1_RPM_MAX, rpm + RPM_MAX_OFFSET);
#ifdef _DEBUG_
			printf("Adjustment down needed (%d-%d). ", rpm, rpm_before);
#endif			
			} else {
#ifdef _DEBUG_
			printf("No adjustment needed (%d-%d). ", rpm, rpm_before);
#endif
			}
			
			/* save rpm value */
			rpm_before = SMCGetFanRpm(SMC_KEY_FAN0_RPM_MAX) - RPM_MAX_OFFSET;
#ifdef _DEBUG_
			printf("Temp = %g, current rpm = %d, target rpm = %d\n", temp, rpm_before, rpm);
#endif
		}
		
		SMCClose();
		
		sleep(10);
	}
	
	return 0;
}
